/* $Id: connection_imp.h,v 2.10 2014/11/30 18:36:27 hurtta Exp $ */

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.10 $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> (was hurtta+elm@posti.FMI.FI)
 *       or Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/


/* service_list.c -------------------------------------------- */

#ifdef REMOTE_MBX

#ifdef I_NETINET_IN
#include <netinet/in.h>
#endif

#ifdef I_ARPA_INET
#include <arpa/inet.h>
#endif

typedef union socket_addr
{
    struct sockaddr         sa;     /* general version */
#ifdef I_NETINET_IN
    struct sockaddr_in      sin;    /* INET family */
#endif
} SOCKADDR;

typedef enum { PORT_end = 0, PORT_smtp = 25, PORT_pop3 = 110, 
	       PORT_imap4 = 143, PORT_submission = 587
} PORTS;

#define    STFLAG_browser     0x01
#define    STFLAG_mbox        0x02
#define    STFLAG_smtp_like   0x04

#define    STFLAG_is_imap   0x0100
#define    STFLAG_is_pop    0x0200
#define    STFLAG_is_submission    0x0400
#define    STFLAG_is_smtp          0x0800

extern const struct service_type * IMAP_SERVICE;
extern const struct service_type * POP_SERVICE;
extern const struct service_type * SUBMISSION_SERVICE;
extern const struct service_type * SMTP_SERVICE;


#define SE_given_alias      (1<<0)
#define SE_given_addr       (1<<1)
#define SE_given_port       (1<<2)
#define SE_temporary        (1<<3)
/* #define SE_unused        (1<<4) */
#define SE_given_name_addr  (1<<5)
#define SE_rescan_name_addr (1<<6)
#define SE_ver_tls_cert     (1<<7)
#define SE_nover_tls_cert   (1<<8)
#define SE_require_tls_name (1<<9)
#define SE_given_name_port  (1<<10)
#define SE_rescan_name_port (1<<11)

#ifdef ANSI_C
struct     SE_option;
#endif

/* return:  0 = failure,
            1 = option ok
*/
typedef int SE_option_parse P_((struct     SE_option *X,
				const char       *name,   /* without prefix */
				const char       * value));

/* return: malloced string -- caller must free */
typedef char * SE_option_values P_((struct     SE_option *X, 
				    const char *prefix));

typedef void SE_option_zero   P_((struct     SE_option *X));
typedef void SE_option_free   P_((struct     SE_option *X));
typedef void SE_option_merge  P_((struct     SE_option *target, 
				  struct     SE_option *source));

/* precious options from target with same name are removed */
typedef void SE_option_opdate  P_((struct     SE_option *target, 
				   struct     SE_option *source));




#define SE_option_t_magic       0xF907
#define SERVICE_ENTRY_magic	0xF90B

extern struct service_entry {
    unsigned short magic;                   /* SERVICE_ENTRY_magic */
    int            refcount;

    int      flags;
    char *   official_name;
    char *   * aliases_list;
    int        aliases_count;
    SOCKADDR * addr_list;
    int        addr_count;

    char    ** addr_name_list;
    int        addr_name_count;

    const struct service_type  * service;

    int      * port_list;
    int        port_count;    

    char    ** port_name_list;
    int        port_name_count;


    struct     SE_option {
	struct SE_option_type {
	    unsigned short magic;    /* SE_option_t_magic */
	    SE_option_parse   * parse_on_option;
	    SE_option_values  * give_values;
	    SE_option_zero    * zero_options;
	    SE_option_free    * free_options;	    
	    SE_option_merge   * merge_options;
	    SE_option_opdate  * update_options;
	} * type;
	char * prefix;

	/* Anonymous type */
	struct SE_option_value  * value;       	
    }        * option_list;
    int        option_count;

    struct string   ** req_tls_peername_list;
    int                req_tls_peername_count;

}  * give_service_entry P_((const char *hostname, int flag));   

extern void free_service_entry P_((struct service_entry **Y));
extern void inc_service_entry_refcount  P_((struct service_entry *Y));

extern int dump_service_entry P_((FILE *f, struct service_entry *entry,
				  charset_t fileset));

extern void free_se_aliases_list P_((struct service_entry * se));
extern void add_alias_to_entry P_((struct service_entry *entry,
				   const char *name));

extern void free_se_addr_list P_((struct service_entry * se));
extern void add_addr_to_entry P_((struct service_entry *entry,
				  SOCKADDR *result));

extern void free_se_addr_name_list P_((struct service_entry * se));
extern void add_addr_name_to_entry P_((struct service_entry *entry,
				       char *name, int malloc_it));

extern void free_se_port_name_list P_((struct service_entry * se));
extern void add_port_name_to_entry P_((struct service_entry *entry,
				       char *name, int malloc_it));

extern void free_se_port_list P_((struct service_entry * se));
extern void add_port_to_entry P_((struct service_entry *entry,
				  int port));

extern int add_option_type_to_entry P_((struct service_entry *entry,
					struct SE_option_type * Y,
					const char * prefix));

extern void free_se_rqtls_pername_list P_((struct service_entry * se));
extern void add_req_peername_to_entry P_((struct service_entry *entry,
					  const struct string * req_peername));

#define MAIL_SERVICES_magic	0xF90C

extern struct mail_services_conf {
    unsigned short magic;     /* MAIL_SERVICES_magic */
    
    struct service_entry ** service_list;
    int                     service_count;
    
} * malloc_mail_services_conf P_((void));

#endif

/* remote_mbx.c ---------------------------------------------- */



#ifdef POLL_METHOD

struct Read_Buffer {
    char      *read_buffer;
    int       read_len;
};
struct Write_Buffer {
    char      *write_buffer;
    int       write_len;
};

extern int ReadFromSocket P_((int fd, struct Read_Buffer *buffer,
			      int wanted));
extern int find_crlf P_((struct Read_Buffer *buffer, int add_null));
extern void zero_Read_Buffer P_((struct Read_Buffer *buffer));
extern void free_Read_Buffer P_((struct Read_Buffer *buffer));
extern void cut_line P_((struct Read_Buffer *buffer, int len));
extern void zero_Write_Buffer P_((struct Write_Buffer *buffer));
extern void free_Write_Buffer P_((struct Write_Buffer *buffer));
extern int WriteToSocket P_((int fd, struct Write_Buffer *buffer));
extern void cut_Write_Buffer P_((struct Write_Buffer *buffer, int n));
extern void add_to_Write_Buffer P_((struct Write_Buffer *buffer,
				    char **str,int l));

#endif


#ifdef REMOTE_MBX


#define REMOTE_ACCOUNT_magic    0xF504

struct remote_account {
    unsigned short magic;    /* REMOTE_ACCOUNT_magic */
    SOCKADDR  hostaddr;

    struct streamsched *stream;
    struct service_entry *service;

    char    * username;
    char    * host;
};

extern void zero_remote_account P_((struct remote_account *ra));
extern void clear_remote_account P_((struct remote_account *ra));

extern int connect_remote_account P_((struct remote_account *ra,
				      int *got,
				      struct service_entry *se,
				      PORTS default_portlist[],
				      PORTS force_port));
extern int split_remote_name P_((const char *name,
				 struct remote_account *X,
				 struct service_entry **se,
				 char **rest,
				 int lookup_flags));
extern int remote_account_OK P_((size_t st_size));

/* Makes new connection to same host */
extern int connect_remote_account_1 P_((struct remote_account *ra,
					const struct remote_account *ra_from));


union ra_verify_peer_cb_data {
    void                    * data;
    struct mailer_info_1    * mailer;
    struct folder_info      * folder;
    struct connection_cache * connection;
};

/* return 1 if cb success, 0 failure
   return -1 no connection
 */
typedef int ra_verify_peer_cb_f P_((struct remote_account *ra,
				    union ra_verify_peer_cb_data *data,
				    const struct string * Server,
				    const struct string * server));

/* cb is called so that possible STARTTLS is started */


/* 1 verify OK
   0 verify failed
  -1 no connection or other failure
*/

extern int remote_account_verify_peer P_((struct remote_account *ra,
					  union ra_verify_peer_cb_data *data,
					  ra_verify_peer_cb_f * cb,
					  const struct string * Server,
					  const struct string * server));



#ifdef ANSI_C
struct connection_cache;
struct folder_browser;
#endif

typedef void  cache_zero_connection P_((struct connection_cache *c));
typedef void  cache_free_connection P_((struct connection_cache *c));

typedef int  cache_open_connection P_((struct connection_cache *c));
typedef int  cache_login_connection P_((struct connection_cache *c,
					const struct string *password /* May be NULL */));
typedef int  cache_close_connection P_((struct connection_cache *c));

typedef void cache_folder_from_connection P_((struct connection_cache *c,
					      struct folder_info *f));
typedef void cache_browser_from_connection P_((struct connection_cache *c,
					       struct folder_browser *d));

typedef struct string * cache_verify_server_const P_((void));

/* return 1 if cb success, 0 failure
   return -1 no connection
 */
typedef int cache_verify_cb_connection P_((struct connection_cache *c,
					   struct remote_account *ra,
					   const struct string * Server,
					   const struct string * server));

#define CONNECTION_TYPE_magic   0xF906

struct connection_type { 
    unsigned short                magic;   /* CONNECTION_TYPE_magic */
    const char                    * type_name;
    cache_zero_connection         * cache_zero_it;
    cache_free_connection         * cache_free_it;
    cache_open_connection         * cache_open_it;
    cache_login_connection        * cache_login_it;
    cache_close_connection        * cache_close_it;      
    cache_folder_from_connection  * cache_folder_from_it;
    cache_browser_from_connection * cache_browser_from_it;
    cache_verify_server_const     * cache_verify_server_const;
    cache_verify_server_const     * cache_verify_Server_const;
    cache_verify_cb_connection    * cache_verify_cb_connection;
};

/* Structure is on connection cache list if not given to
   folder - structure ...
*/

enum connection_state { 
    CON_greeting = -1,
    CON_error = 0, CON_open = 1, CON_logged = 2,

    CON_verify_failure = -2
 };

struct connection_cache {
    struct connection_type  * type;

    /* Either one from these only should be on non-NULL */
    struct connection_cache * next;       
    struct folder_info      *f;
    struct folder_browser   *d;

    struct remote_account   C;
    struct remote_server   * this_server;  

    uint16                  port;    /* 0 == default port */
    enum connection_state   state;

    union {
	void                         * any;
	struct generic_connection_1  * g1_con;
	struct generic_connection_2  * g2_con;

        struct IMAP_CON              * imap_con;
    } a;
};

/* May not be 
      uin16 port
   because this is promoted to int
*/


extern struct connection_cache * 
locate_from_cache P_((const char * username,
		      const char * host,
		      const struct connection_type *con_type,
		      int port    /* 0 == default port */));


extern struct connection_cache * create_connection P_((struct connection_type *T));
extern void  free_connection P_((struct connection_cache **c));

extern int  join_connection P_((struct connection_cache *c,
				struct remote_account *X,
				enum connection_state st));

extern int  login_connection P_((struct connection_cache *c,
				 const struct string *password /* May be NULL */));

extern void folder_from_connection P_((struct connection_cache *c,
				       struct folder_info *f));

extern void browser_from_connection P_((struct connection_cache *c,
					struct folder_browser *d));

extern void  close_connection P_((struct connection_cache *c));

extern void  cache_connection P_((struct connection_cache *c));

#endif

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */

